#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 5366 $
$LastChangedDate:: 2021-12-20 17:55:34 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import sys
import json
import optparse
import shutil

sym_root = os.getenv('MagOpt_Home')
cs_root = os.getenv('ComponentSpace')

if sym_root is None:
    raise EnvironmentError('MagOpt_Home is not defined!')
elif cs_root is None:
    raise EnvironmentError('ComponentSpace is not defined!')
else:
    sys_paths = [os.path.join(sym_root, 'System', 'Python', 'src'),
                 os.path.join(sym_root, 'System', 'General', 'Python', 'src'),
                 os.path.join(cs_root, 'src')]

    for path in sys_paths:
        if path not in sys.path:
            sys.path.append(path)

    from symspace.osutils.pathutils import Paths
    from symspace.osutils.pathutils import CMDParser
    from symspace.testing.symproject.case import DefaultKwargs
    from symspace.testing.symproject.case import SymProjectTestCase
    from symspace.testing.symproject.suite import SymProjectTestSuite
    from symspace.osutils.jenkins_utils import emulateBuildServer
    from componentspace.create_buildscript import ComponentBuilder
    from componentspace.testing.create_tests import ComponentTestBuilder
    from componentspace.testing.project_tests import ProjectTestBuilder
    from componentspace.component_management import ComponentManager


def runFromCMD():
    parser = optparse.OptionParser("usage: %prog [options]")
    parser.add_option("-o", "--function", dest="function",
                      type="string",
                      help="Function to run")
    parser.add_option("-f", "--file", dest="json_file",
                      type="string",
                      help="Json file with list of selected components")
    parser.add_option("-e", "--emptyUserGroups", dest="empty_usergroups",
                      default=0, type="int",
                      help="delete existing components from usergroups folder, default is 0 (False)")
    parser.add_option("--build_creo", dest="build_creo",
                      default=0, type="int",
                      help="build creo tests, default is 0 (False)")
    parser.add_option("--build_freecad", dest="build_freecad",
                      default=0, type="int",
                      help="build freecad tests, default is 0 (False)")
    parser.add_option("--run_sim", dest="run_sim",
                      default=0, type="int",
                      help="set run simulation flag, default is 0 (False)")
    parser.add_option("--remote_compute", dest="remote_compute",
                      default=0, type="int",
                      help="set remote compute flag, default is 0 (False)")
    parser.add_option("--timeout", dest="timeout",
                      default=60.0, type="float",
                      help="process timeout, default is 60")
    parser.add_option("--major", dest="major",
                      default=0, type="int",
                      help="set major version, default is 9")
    parser.add_option("--minor", dest="minor",
                      default=0, type="int",
                      help="set minor version, default is 0")
    parser.add_option("--increase_version", dest="increase_version",
                      default=60.0, type="int",
                      help="If True: Increase old version by major.minor; "
                           "Else: Replace old version by major.minor, default is 0 (False)")
    parser.add_option("--pkg_file", dest="pkg_json_file",
                      default='', type="string",
                      help="Json file with list of package")
    parser.add_option("--auto_collect_pkgs", dest="auto_collect_pkgs",
                      default=0, type="int",
                      help="Automatically collect the package dependencies")
    parser.add_option("--emulate_build_server", dest="emulate_build_server",
                      default=0, type="int",
                      help="Emulate the build server")
    parser.add_option("--n_max_proc", dest="n_max_proc",
                      default=1, type="int",
                      help="Number of max. parallel processes")

    args = CMDParser.jepSysArgs()
    (options, args) = parser.parse_args(args)
    if len(args) != 0:
        parser.error("incorrect number of arguments")

    emulateBuildServer(bool(options.emulate_build_server))

    if options.function == 'build_components':
        empty_usergroups = bool(options.empty_usergroups)
        json_file = CMDParser.jepArgPath(options.json_file)
        build_components(json_file, empty_usergroups)
    elif options.function == 'build_component_tests':
        json_file = CMDParser.jepArgPath(options.json_file)
        build_creo = bool(options.build_creo)
        build_freecad = bool(options.build_freecad)
        build_component_tests(json_file, build_creo=build_creo, build_freecad=build_freecad)
    elif options.function == 'run_component_tests':
        json_file = CMDParser.jepArgPath(options.json_file)
        run_sim = bool(options.run_sim)
        remote_compute = bool(options.remote_compute)
        timeout = float(options.timeout)
        n_max_proc = int(options.n_max_proc)
        run_component_tests(json_file, run_sim, remote_compute, timeout, n_max_proc)
    elif options.function == 'build_excel_tests':
        json_file = CMDParser.jepArgPath(options.json_file)
        build_excel_tests(json_file)
    elif options.function == 'run_excel_tests':
        json_file = CMDParser.jepArgPath(options.json_file)
        run_sim = bool(options.run_sim)
        remote_compute = bool(options.remote_compute)
        timeout = float(options.timeout)
        run_excel_tests(json_file, run_sim, remote_compute, timeout)
    elif options.function == 'collect_info':
        json_file = CMDParser.jepArgPath(options.json_file)
        collect_info(json_file)
    elif options.function == 'change_version':
        json_file = CMDParser.jepArgPath(options.json_file)
        major = int(options.major)
        minor = int(options.minor)
        increase_version = bool(options.increase_version)
        change_version(json_file, major, minor, increase_version)
    elif options.function == 'add_packages':
        json_file = CMDParser.jepArgPath(options.json_file)
        pkg_json_file = CMDParser.jepArgPath(options.pkg_json_file)
        auto_collect_pkgs = bool(options.auto_collect_pkgs)
        add_packages(json_file, pkg_json_file, auto_collect_pkgs)
    else:
        raise NotImplementedError('Unexpected case!')
    print('Finished')


def build_components(json_file: str, empty_usergroups: bool):
    components_list = json.loads(open(json_file).read())

    builder = ComponentBuilder(start_path=Paths.getComponentSpaceRoot(), full_build=False)

    if empty_usergroups:
        builder.emptyUserGroups()

    sorted_components_list = builder.sortComponentslist(components_list)
    builder.buildComponents(sorted_components_list)
    builder.deployReports()


def build_component_tests(json_file: str, build_creo: bool, build_freecad: bool):
    components_list = json.loads(open(json_file).read())

    test_builder = ComponentTestBuilder(start_path=Paths.getComponentSpaceRoot(), full_build=True)

    test_builder.build_creo = build_creo
    test_builder.build_freecad = build_freecad

    test_builder.cleanReportsDir()
    test_builder.buildTests(components_list)
    test_builder.deployReports()


def run_component_tests(json_file: str, run_sim: bool, remote_compute: bool, timeout, n_max_proc: int):
    components_list = json.loads(open(json_file).read())

    report_dir = ComponentTestBuilder.run_reports_dir
    output = os.path.join(report_dir, 'single_tests')
    merged_file = os.path.join(report_dir, 'merged-junit.xml')

    if os.path.isdir(report_dir):
        shutil.rmtree(report_dir)

    test_case_options = SymProjectTestCase.get_test_case_options(RunSim=run_sim,
                                                                 RemoteCompute=remote_compute,
                                                                 Optimize=False,
                                                                 Timeout=timeout,
                                                                 skip_preview=True,
                                                                 skip_docfiles=True,
                                                                 save_finished=False,
                                                                 fail_incomplete=True,
                                                                 stop_on_error=True,
                                                                 )

    suite = SymProjectTestSuite()

    prj_files = []
    for directory in components_list:
        prj_files.extend(suite.findTestsFilesByPattern(directory, ComponentTestBuilder.get_search_pattern()))

    suite.addTestsFromFiles(prj_files, **test_case_options)
    suite.runExternalXMLRunner(output, timeout=timeout, n_max_proc=n_max_proc)
    suite.mergeJUnitXMLFiles(merged_file, output=output)


def build_excel_tests(json_file: str):
    file_list = json.loads(open(json_file).read())
    builder = ProjectTestBuilder(start_path=Paths.getComponentSpaceRoot())
    builder.cleanReportsDir()
    builder.buildTests(file_list)
    builder.deployReports()


def run_excel_tests(json_file: str, run_sim: bool, remote_compute: bool, timeout):
    files_list = json.loads(open(json_file).read())

    report_dir = ProjectTestBuilder.run_reports_dir
    output = os.path.join(report_dir, 'single_tests')
    merged_file = os.path.join(report_dir, 'merged-junit.xml')

    if os.path.isdir(report_dir):
        shutil.rmtree(report_dir)

    test_case_options = SymProjectTestCase.get_test_case_options(RunSim=run_sim,
                                                                 RemoteCompute=remote_compute,
                                                                 Optimize=False,
                                                                 Timeout=timeout,
                                                                 )

    suite = SymProjectTestSuite()

    prj_files = []
    for xls_file in files_list:
        directory = xls_file.replace(ProjectTestBuilder.xls_file_ending, '')
        prj_files.extend(suite.findTestsFilesByPattern(directory, ProjectTestBuilder.get_search_pattern()))

    suite.addTestsFromFiles(prj_files, **test_case_options)
    suite.runExternalXMLRunner(output, timeout=timeout)
    suite.mergeJUnitXMLFiles(merged_file, output=output)


def collect_info(json_file: str):
    components_list = json.loads(open(json_file).read())
    manager = ComponentManager(start_path=Paths.getComponentSpaceRoot())
    manager.collectInfo(components_list=components_list)


def change_version(json_file: str, major: int, minor: int, increase_version: bool):
    components_list = json.loads(open(json_file).read())
    manager = ComponentManager(start_path=Paths.getComponentSpaceRoot())
    manager.changeVersion(major, minor, increase_version, components_list=components_list)


def add_packages(json_file: str, pkg_json_file: str, auto_collect: bool):
    components_list = json.loads(open(json_file).read())
    packages = json.loads(open(pkg_json_file).read())
    manager = ComponentManager(start_path=Paths.getComponentSpaceRoot())
    manager.addPackages(packages, auto_collect=auto_collect, components_list=components_list)


if __name__ == '__main__':
    runFromCMD()
