#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 5322 $
$LastChangedDate:: 2021-12-06 15:31:58 +0100#$

This file is part of SyMSpace.

Copyright (c) 2019, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import numpy as np
from symspace.wrapper.symwrapper import SymProject
from symspace.gui.qt.form_dialog import FormDialog


class CompareProject:

    def __init__(self, src_prj, dst_prj):
        self.src_prj = src_prj
        self.dst_prj = dst_prj

    @staticmethod
    def _isEqual(src, dst):
        """
        Compare parameters if they are equal

        Parameters
        ----------
        src : SymParameter
        dst : SymParameter

        Returns
        -------
        bool
        """
        ret = False
        if isinstance(dst.getData(), np.ndarray):
            if np.array_equal(src.getData(), dst.getData()):
                ret = True
        elif src.getData() == dst.getData():
            ret = True
        return ret

    @staticmethod
    def _getScore(src, dst):
        lst_src = src.getIdentifier().split('.')[::-1]
        lst_dst = dst.getIdentifier().split('.')[::-1]
        score = 0
        for k in range(min((len(lst_dst), len(lst_src)))):
            if lst_src[k] == lst_dst[k]:
                score += 1
            else:
                break
        return score

    def getScores(self, param):
        """
        Get scores of parameter with name in source and destination project

        Parameters
        ----------
        param : SymParameter
            Parameter

        Returns
        -------
        ret : tuple
            (source_parameter, destination_parameter)
        """
        name = param.getName()
        score = []
        par_lst = []
        for sp in self.src_prj.getParameters('SymParameter'):
            if sp.getName() == name:
                for dp in self.dst_prj.getParameters('SymParameter'):
                    if dp.getName() == name:
                        score.append(self._getScore(sp, dp))
                        par_lst.append((sp, dp))
        if score.count(max(score)) == 1:
            idx = score.index(max(score))
            return [par_lst[idx][1]]
        else:
            ret = []
            ret_score = []
            for i in range(len(score)):
                if score[i] > 1:
                    if param in par_lst[i]:
                        ret.append(par_lst[i][1])
                        ret_score.append(score[i])
            return ret

    def matchParam(self):
        """
        Match parameters in source and destination project
        """

        match_lst = []

        src_param = self.src_prj.getParameters('SymParameter')
        src_names = [p.getName() for p in src_param]
        dst_param = self.dst_prj.getParameters('SymParameter')
        dst_names = [p.getName() for p in dst_param]

        src_unused = []
        dst_unused = []
        for param in src_param:
            cnt = dst_names.count(param.getName())
            if cnt == 1 and src_names.count(param.getName()) == 1:
                idx = dst_names.index(param.getName())
                match_lst.append((param, dst_param[idx]))
            elif cnt >= 1:
                match = self.getScores(param)
                if len(match) == 1:
                    match_lst.append((param, match[0]))
                else:
                    src_unused.append(param)
            else:
                src_unused.append(param)

        # get unused dst parameters
        dst_lst = [p[1] for p in match_lst]
        for param in dst_param:
            if param not in dst_lst:
                dst_unused.append(param)
            elif dst_lst.count(param) > 1:
                print('Found multiple {}'.format(param.getIdentifier()))
                for item in match_lst:
                    if item[1] == param:
                        print('    {} --> {} : Score {}'.format(item[0].getIdentifier(), item[1].getIdentifier(),
                                                                self._getScore(item[0], item[1])))
                print('\n')

        return match_lst, src_unused, dst_unused

    def updateParam(self):
        """
        Updates parameters from source project if parameters are equivalent
        """
        file = os.path.join(self.dst_prj.getDirectory(), 'update.log')
        fh = open(file, 'w')
        fh.write('Copy parameters: {} --> {} \n\n'.format(self.src_prj.getFile(), self.dst_prj.getFile()))

        dst_inparam = self.dst_prj.getUsedInputParameter()
        match, src_unused, dst_unused = self.matchParam()
        dst_lst = [p[1] for p in match]
        dst_unused = []

        for par in dst_inparam:
            info = [par.getIdentifier()]
            if par in dst_lst:

                idx = dst_lst.index(par)
                src_par = match[idx][0]
                if par.getLock():
                    info.append('Data is locked')
                elif self._isEqual(par, src_par):
                    info.append('Data is equal')
                else:
                    info.append('Source: {}'.format(src_par.getIdentifier()))
                    info.append('new value {} instead of {}'.format(str(src_par.getData()), str(par.getData())))
                    par.setData(src_par.getData())
            else:
                dst_unused.append(par)
                info = []

            if info:
                fh.write(', '.join(info))
                fh.write('\n')

        fh.write('\n\n')
        for par in dst_unused:
            fh.write('Not used: {}\n'.format(par.getIdentifier()))
        fh.close()

    def compareParam(self):
        """
        Compares two projects and writes csv file
        """
        file = os.path.join(self.dst_prj.getDirectory(), 'compare.csv')
        with open(file, 'w') as fh:
            fh.write('Compare projects\n\n')
            fh.write('Source:; {}\nDestination:; {}\n\n'.format(self.src_prj.getFile(), self.dst_prj.getFile()))
            match, src_unused, dst_unused = self.matchParam()
            dst_lst = [p[1] for p in match]

            for par in self.dst_prj.getParameters('SymParameter'):
                fh.write(' ' * par.getIdentifier().count('.') + par.getName() +
                         ';{};'.format(par.getIdentifier().replace(self.dst_prj.getName() + '.', '')))
                if par in dst_lst:
                    mparam = match[dst_lst.index(par)]
                    fh.write('{}; {}; {}'.format(mparam[0].getIdentifier().replace(self.src_prj.getName() + '.', ''),
                                                 mparam[1].getDataString().replace('.', ','),
                                                 mparam[0].getDataString().replace('.', ',')))
                fh.write('\n')

            fh.write('\n\n')

            fh.write('Unused source parameters\n')
            for par in src_unused:
                fh.write('{}\n'.format(par.getIdentifier()))


if __name__ == '__main__':

    __project__ = locals()['__project__']
    form_title = 'Open Source File'
    form_elements = [('Select SyMSpace file', 'browse_file', '', 'Project files (*.mop)'),
                     ('Action', 'combobox', ['Copy parameters', 'Compare project'])]
    form_result = FormDialog.run(form_title, form_elements)
    
    if form_result['accept']:
        _src_prj = form_result['Select SyMSpace file']
        print('Source project: {}'.format(_src_prj))
        print('Action: {}'.format(form_result['Action']))
        prj = SymProject.load(_src_prj)
        update = CompareProject(prj, __project__)
        if form_result['Action'] == 'Copy parameters':
            update.updateParam()
            print('Copy finished')
        else:
            update.compareParam()
            print('Compare finished')

    else:
        print('Cancelled by user')
