#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 5279 $
$LastChangedDate:: 2021-11-26 09:35:30 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import glob
import datetime
import json
import optparse
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
from symspace.osutils.pathutils import Paths
from symspace.osutils.pathutils import CMDParser


class ComponentDependencyCollector:
    """Class to collect the dependency of components to other components"""

    @classmethod
    def fromCMDParser(cls) -> 'ComponentDependencyCollector':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        --libraryPath : str
            Search in componentspace or in symspace for mopc-files, componentspace (Default)"

        Returns
        -------
        collector : ComponentDependencyCollector
        """

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-p", "--libraryPath", dest="library_path",
                          default='componentspace', type="str",
                          help="Search in componentspace or in symspace for mopc-files, componentspace (Default)")

        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)

        library_path = options.library_path

        print('libraryPath: {}'.format(library_path))

        collector = cls(library_path)
        return collector

    def __init__(self, library_path='componentspace'):
        """
        Initialize class for simple use

        Parameters
        ----------
        library_path : str
            Search in ComponentSpace (Default) or in SyMSpace for json-files
        """
        if library_path.lower() == 'componentspace':
            start_path = os.path.join(Paths.getComponentSpaceRoot(), '_UserGroups')
        else:
            start_path = os.path.join(Paths.getSyMSpaceRoot(), 'Library', '_ComponentSpace')

        self.__start_path = start_path
        self.__cs_jenkins = 'C:\\opt\\Jenkins\\jobs\\ComponentSpace\\workspace\\ComponentSpace'

    def setStartPath(self, start_path: str):
        """Change the start path (e.g. for debugging)"""
        self.__start_path = start_path

    def update(self, plot_graph: bool = False):
        """
        Update the dependencies

        Parameters
        ----------
        plot_graph: bool
            Plot the dependency graph, False (Default)

        """
        folder = os.path.join(Paths.getComponentSpaceRoot(), 'dependencies')
        json_error_file = os.path.join(folder, 'error_dependencies.json')
        json_file = os.path.join(folder, 'build_dependencies.json')
        xls_file = os.path.join(folder, 'build_dependencies.xlsx')

        error_collect = []
        error_map = []

        d_paths = self.__collectDependencies(self.__start_path, error_list=error_collect, error_map=error_map)
        d_errors = dict(error_collect=error_collect, error_map=error_map)
        d_paths = self.__convertAbsolutToRelative(d_paths)

        os.makedirs(folder, exist_ok=True)
        self.__exportDictToJSON(d_paths, json_file)
        self.__exportDictToJSON(d_errors, json_error_file)
        self.__exportDictToXLS(d_paths, xls_file)

        if plot_graph:
            self.__plotGraph(d_paths)

    def __collectDependencies(self, start_path: str, print_summary: bool = True, error_list: list = None,
                              error_map: list = None):
        """
        Collect dependencies from json files below the start_path

        Parameters
        ----------
        start_path: str
            Start search of json-files from this folder
        print_summary: bool
            Print a error statistic, True (Default)
        error_list: list
            Return failures in this list, None (Default)

        """
        d_webpaths = {}
        d_dirpaths = {}
        d_mapping = {}
        cnt_failed_main = 0
        cnt_failed_map = 0

        if error_list is None:
            error_list = []
        if error_map is None:
            error_map = []

        filenames = glob.glob(start_path + '/**/*.json', recursive=True)
        n_components = len(list(filenames))
        print('\nFound {} components'.format(n_components))

        for index, json_file in enumerate(filenames):
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            disp_path = os.path.basename(json_file).replace('.json', '')
            print('{} Collect dependencies {}/{}: {}'.format(timestamp, index + 1, n_components, disp_path))
            try:
                main_wp, main_dir, deps_wp = self.__collectWebpaths(json_file)

                '''only add unique dependencies'''
                if main_wp in d_webpaths:
                    d_webpaths[main_wp].extend(deps_wp)
                else:
                    d_webpaths[main_wp] = deps_wp
                    d_mapping[main_wp] = main_dir

            except Exception as ex:
                error_list.append(disp_path)
                cnt_failed_main += 1
                print('\t=> Failed: {}'.format(str(ex)))

        n_main = len(d_webpaths.keys())
        for index, _main_wp in enumerate(d_webpaths.keys()):
            _main_path = d_mapping[_main_wp]
            if _main_path not in d_dirpaths:
                d_dirpaths[_main_path] = []
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Map dependencies {}/{}: {}'.format(timestamp, index + 1, n_main, _main_wp))
            for _dep_wp in d_webpaths[_main_wp]:
                try:
                    if _dep_wp not in d_mapping:
                        raise KeyError('Webpath: {} not in Main Webpaths!'.format(_dep_wp))
                    _dep_path = d_mapping[_dep_wp]
                    d_dirpaths[_main_path].append(_dep_path)
                except Exception as ex:
                    error_map.append(_dep_wp)
                    cnt_failed_map += 1
                    print('\t=> Failed: {}'.format(str(ex)))
                    
            '''remove duplicates'''
            d_dirpaths[_main_path] = list(set(d_dirpaths[_main_path]))
            '''sort for easier comparison'''
            d_dirpaths[_main_path].sort()

        if print_summary:
            print('\nCollect Summmary')
            print('----------------')
            print('Failed Main: {}/{}'.format(cnt_failed_main, n_components))
            print('Failed Mapping: {}'.format(cnt_failed_map, n_components))
            print('Failed Total: {}\n'.format(cnt_failed_main + cnt_failed_map))

        return d_dirpaths

    @staticmethod
    def __convertAbsolutToRelative(abs_dependencies: dict) -> dict:
        """
        Replace absolute path with relative placeholders of a complete dictionary

        Parameters
        ----------
        abs_dependencies: dict
            Dictionary with absolute paths

        Returns
        -------
        rel_dependencies : dict
            Dictionary with relative paths
        """

        def toRelPath(abs_val: str) -> str:
            """replace absolute path with relative placeholders"""
            return abs_val.replace(Paths.getComponentSpaceRoot(), '%ComponentSpace%')

        rel_dependencies = {}

        for abs_key in abs_dependencies.keys():
            '''convert all absolute to relative paths'''
            rel_key = toRelPath(abs_key)
            rel_dependencies[rel_key] = []
            for abs_dep in abs_dependencies[abs_key]:
                rel_dep = toRelPath(abs_dep)
                rel_dependencies[rel_key].append(rel_dep)

        return rel_dependencies

    @staticmethod
    def __reduceNames(abs_dependencies: dict) -> dict:
        """
        Reduce the names of a complete dictionary to the last' .' or '\\' separated element

        Can be useful when plotting a graph

        Parameters
        ----------
        abs_dependencies : dict
            Dictionary with full names

        Returns
        -------
        red_dependencies : dict
            Dictionary with reduced names
        """

        def toRedPath(abs_val: str) -> str:
            """replace absolute path with relative placeholders"""
            red_val = abs_val.split('.')[-1]
            red_val = red_val.split('\\')[-1]
            return red_val

        red_dependencies = {}

        for abs_key in abs_dependencies.keys():
            '''convert all absolute to relative paths'''
            rel_key = toRedPath(abs_key)
            red_dependencies[rel_key] = []
            for abs_dep in abs_dependencies[abs_key]:
                rel_dep = toRedPath(abs_dep)
                red_dependencies[rel_key].append(rel_dep)

        return red_dependencies

    @staticmethod
    def __collectWebpaths(json_file: str) -> tuple:
        """
        Collect component webpaths from a json-file

        Parameters
        ----------
        json_file : str

        Returns
        -------
        tuple
        """

        if not os.path.isfile(json_file):
            raise FileNotFoundError(json_file)

        d = json.loads(open(json_file).read())
        main_webpath = d['Info']['WebPath']
        main_dir = d['Dependencies']['base_directory']
        webpaths = d['Dependencies']['webpaths']

        build_script = os.path.join(main_dir.replace('%ComponentSpace%', Paths.getComponentSpaceRoot()),
                                    'build_component.py')

        if not os.path.isfile(build_script):
            raise FileNotFoundError('Component folder not found! {}'.format(main_webpath))

        return main_webpath, main_dir, webpaths

    @staticmethod
    def __exportDictToXLS(d_in: dict, file_path: str):
        """
        Export dictionaries to excel-file

        Parameters
        ----------
        d_in: dict
            Dictionary to export
        file_path : str
            Filepath to export to
        """
        d_out = {}
        for key, value in d_in.items():
            d_out[key] = (value,)

        df = pd.DataFrame.from_dict(d_out).transpose()
        df.to_excel(file_path)
        print('File save to: {}'.format(file_path))

    @staticmethod
    def __exportDictToJSON(d_in: dict, file_path: str):
        """
        Export dictionaries to json-file

        Parameters
        ----------
        d_in: dict
            Dictionary to export
        file_path : str
            Filepath to export to
        """
        timestamp = str(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
        py_file = os.path.abspath(__file__).replace(Paths.getComponentSpaceRoot(), '%ComponentSpace%')
        d_in['!COMMENT!'] = '{} Automatically generated file by: {}'.format(timestamp, py_file)
        with open(file_path, 'w') as fh:
            json.dump(d_in, fh, indent=4, sort_keys=True)

        print('File save to: {}'.format(file_path))

    @classmethod
    def __plotGraph(cls, d: dict):
        """
        Plot a dependencies Graph of a dictionary

        Parameters
        ----------
        d: dict
            Dictionary with dependency paths
        """

        d_red = cls.__reduceNames(d)
        # g = nx.Graph(d_red)
        # g = nx.DiGraph(d_red)
        # g = nx.MultiGraph(d_red)
        g = nx.MultiDiGraph(d_red)

        plt.subplot(221)
        nx.draw(g, with_labels=True)
        plt.subplot(222)
        nx.draw_shell(g, with_labels=True)
        plt.subplot(223)
        nx.draw_spring(g, with_labels=True)
        plt.subplot(224)
        # nx.draw_kamada_kawai(g, with_labels=True)
        # nx.draw_spectral(g, with_labels=True)
        nx.draw_networkx(g, with_labels=True)


if __name__ == '__main__':
    cdc = ComponentDependencyCollector(library_path='symspace')
    cdc.update()
