# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 4560 $
$LastChangedDate:: 2021-07-13 16:56:47 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import asyncio
import logging
import json_tricks
import random
import signal
import sys
import time

from neuropil import NeuropilNode, np_token, np_message, NeuropilException
from _neuropil import lib as np

logging.basicConfig(filename='neuropil_client.log', format='%(asctime)s %(name)-15.15s:%(levelname)-7.7s : %(message)s',
                    datefmt='%Y:%m:%d %H:%M:%S', level=logging.DEBUG)

logger = logging.getLogger('neuropil_client')
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel(logging.DEBUG)


class NeuropilClient(NeuropilNode):
    """
    example neuropil node to calculate the square of two big numbers using numpy
    result will be send back on the same subject
    """

    def __init__(self, settings: dict):
        super(NeuropilClient, self).__init__(port=random.randint(1025, 10250), proto=b'pas4',
                                             log_file="./neuropil_client_tmp.log", n_threads=0)
        self.done = False
        self.subject = settings['subject']
        self.bootstrap_node = settings['bootstrap_node']
        self.replies = settings['replies']
        self.result = asyncio.Queue()

        logger.info("requesting to join {bt_host} network ...".format(bt_host=self.bootstrap_node))
        self.join(self.bootstrap_node)

        self.set_authorize_cb(self.authorize)
        self.set_authenticate_cb(self.authenticate)

        while not self.has_joined():
            logger.info("waiting to join the network ...")
            time.sleep(0.1)

        logger.info("successfully joined the network as: {node_hash}".format(node_hash=str(self.get_fingerprint())))

        properties = self.get_mx_properties(self.subject)
        properties.ackmode = np.NP_MX_ACK_CLIENT
        # properties.intent_update_after = 10.0
        properties.max_retry = 3
        properties.apply()

        properties = self.get_mx_properties(self.replies)
        properties.ackmode = np.NP_MX_ACK_CLIENT
        # properties.intent_update_after = 10.0
        properties.max_retry = 3
        properties.apply()

        logger.info('waiting on {subject} for incoming data'.format(subject=self.replies))
        self.set_receive_cb(self.replies, self.run_fnc)
        logger.info('setup done: {connect_info}'.format(connect_info=self.bootstrap_node))

    def handle_reconnect(self):
        self.join(self.bootstrap_node)

    def handle_shutdown(self, signum, frame):
        self.shutdown()

    @staticmethod
    def authorize(self, token: np_token):
        logger.info('authorized access for token {token}'.format(token=token.get_fingerprint()))
        return True

    @staticmethod
    def authenticate(self, token: np_token):
        logger.info('authenticated access for token {token}'.format(token=token.get_fingerprint()))
        return True

    async def start(self):
        # i = 0
        status = np.np_ok
        result = None
        logger.info(f"Starting event loop")
        try:
            while not self.done:
                # while status == np.np_ok:
                # while status == np.np_ok and result is None:
                # logger.debug("running client event loop: {iteration} | {hash}".format(iteration=i, hash=self.get_fingerprint()))
                status = self.run(0)
                # i = i + 1
                # if (i % 100) == 0:
                #     logger.info('{} {} {}'.format(status, type(status), type(np.np_ok)))
                # if i == 1000:
                #     status = 1
                await asyncio.sleep(0)
                # result = await self.result.get()
        except NeuropilException:
            logger.info("stopping neuropil main event loop")
        finally:
            logger.info(f"stopping event loop")
            self.shutdown()
            return result

    async def getResult(self):
        logger.info("waiting for result ...")
        result = await self.result.get()
        logger.info("got result ...")
        self.shutdown()
        logger.info('Received result: {}'.format(result))
        self.done = True
        return result

    async def send_data(self, input_data):
        i = 0
        while not self.np_has_receiver_for(self.subject):
            logger.debug("waiting for receiver: {iteration} | {hash}".format(iteration=i, hash=self.get_fingerprint()))
            await asyncio.sleep(1.0)
            i = i + 1
        logger.info("sending data ...")
        self.send(self.subject, input_data)

    def run_fnc(self, input_message: np_message):
        logger.info(
            'received new result data for subject {subject}'.format(subject=self.replies))
        self.result.put_nowait(input_message.raw().decode('utf-8'))

        return True

    def run_loop(self, input_data: dict):
        logger.info("installing shutdown hook")
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, self.handle_shutdown)
        loop.add_signal_handler(signal.SIGTERM, self.handle_shutdown)
        loop.add_signal_handler(signal.SIGHUP, self.handle_reconnect)

        _, _, result_json = loop.run_until_complete(asyncio.gather(self.start(),
                                                                   self.send_data(
                                                                       bytes(json_tricks.dumps(input_data),
                                                                             encoding='UTF-8')),
                                                                   self.getResult()
                                                                   ))
        # _, result = loop.run_until_complete(asyncio.gather(self.start(),
        #                                                    self.send_data(
        #                                                        bytes(json_tricks.dumps(input_data), encoding='UTF-8'))
        #                                                    ))
        result = json_tricks.loads(result_json)

        logger.info('Received result type: {}'.format(type(result)))
        logger.info('client run_loop() finished')
        return result


if __name__ == '__main__':
    _settings = {
        # data dependent configuration / information
        'namespace': b'at:lcm:symspace',
        'subject': b'at:lcm:symspace:a_square_b_square',
        'datamodel_in': None,
        'replies': b'at:lcm:symspace:a_square_b_square:reply',
        'datamodel_out': None,

        # technical connectivity details
        'bootstrap_node': '*:udp4:localhost:2345',
        # 'bootstrap_node': '*:udp4:demo.neuropil.io:3400',

    }
    _input_data = dict(a=3, b=5)
    _app = NeuropilClient(_settings)
    _output_data = _app.run_loop(_input_data)
